page 60,132
Title   Example Zoomer PDA serial code
.model tiny
LOCALS
subttl PenRight! Zoomer PDA Serial Port Sample Program
;********************************************************************
;* PenRight! Zoomer PDA Serial Port Sample Program
;*
;* Copyright 1993,1994 by PenRight! Corporation.
;*
;*
;* The interface to the serial port of the Zoomer PDA is not compatible with
;* the standard 8250 UART.  This program provides a working sample which
;* supports interrupt driven transmitting and receiving.
;*
;* The interrupt controller in the Zoomer PDA is different as well.
;*
;* The beginning of the program is mostly a description of the various
;* I/O ports used. 
;*
;* A queue is currently used for transmitting, and a queue is implemented
;* for receiving, although the receive queue has not been tested.
;*
;* Further note that the line status interrupts and modem status interrupts
;* are not supported, nor is flow control or anything else.  This is just
;* supposed to be a sample program.
;*
;* It was tested with a GRiD 2390 (Zoomer PDA) connected to a GRiD 1550
;* running Qmodem.  Any other communications program should work fine.
;*
;* This sample code does not support PCMCIA serial-type adapters.
;* The driver for a PCMCIA card would require that the PCMCIA slot
;* and card be configured first, and all of the code talking directly to
;* the UART would have to be changed (xmit_char, the initialization code, and
;* the interrupt handler).   The PCMCIA goes through IRQ 8 (not tested),
;* which is mapped to interrupt 78h respectively.
;*
;* The UART used in the Zoomer PDA offers most (if not all) of the features
;* of an 8250 or 16450 UART, the standard UARTs used in IBM PC's.
;*
;* The interrupt controller in the Zoomer PDA is not compatible to the one
;* used in the PC.  However, programming the interrupt controller is pretty
;* straight forward.
;*
;*
page
subttl  Baud rate register  (R/W)
;**********************************************
;* Baud rate counter register (R/W), port 50h *
;**********************************************

BAUD_RATE_REG   EQU 50h ; baud rate register definition

; The baud rate is controlled by the baud rate divisor register, 50h
; The register is set up as follows:
;
;   D7  D6  D5  D4  D3  D2  D1  D0
;    - BR2 BR1 BR0 PR3 PR2 PR1 PR0
;
;       BR              PR
;
;   BR2  BR1  BR0  |  BR        PR3  PR2  PR1  PR0  |  PR
;   ---------------+-------     --------------------+------
;    0    0    0   |  1/2        0    0    0    0   |  1/1
;    0    0    1   |  1/4        0    0    0    1   |  1/2
;    0    1    0   |  1/8        0    0    1    0   |  1/3
;    0    1    1   |  1/16       0    0    1    1   |  1/4
;    1    0    0   |  1/32       0    1    0    0   |  1/5
;    1    0    1   |  1/64          .       .
;    1    1    0   |  1/128         .       .
;    1    1    1   |  1/256         .       .
;                    1    1    1    1   |  1/16
;
; The baud rate may be calculated with the following equation:
;   Baud Rate = PR*BR*Fosc/32   Fosc = 7.3728MHz    Fosc/32 = 230400
;
;   Example:  Baudrate = 2400 = 1/32 * 1/3 * 230400 so the baud rate
;       counter register would be set to 42h
;
; The table below contains values for frequently used baud rates
; Please note that not all of the baud rates have been tested and
; that baud rates higher than 38400 are not recomended.

;Zoomer PDA BAUD RATE TABLE
Z115200_BAUD    EQU 000h    ; 115200 baud   (NOT RECOMENDED, slightly tested)
Z57600_BAUD EQU 001h    ; 57600 baud    (NOT RECOMENDED, slightly tested)
Z38400_BAUD EQU 002h    ; 38400 baud
Z19200_BAUD EQU 012h    ; 19200 baud
Z14400_BAUD EQU 021h    ; 14400 baud    (not tested)
Z9600_BAUD  EQU 022h    ; 9600 baud
Z7200_BAUD  EQU 031h    ; 7200 baud
Z4800_BAUD  EQU 032h    ; 4800 baud
Z3600_BAUD  EQU 031h    ; 3600 baud (not tested)
Z2400_BAUD  EQU 042h    ; 2400 baud
Z1200_BAUD  EQU 052h    ; 1200 baud
Z600_BAUD   EQU 062h    ; 600 baud  (not tested)
Z300_BAUD   EQU 072h    ; 300 baud
Z150_BAUD   EQU 075h    ; 150 baud  (not tested)
;
; Theoretically higher baud rates can be set, but the hardware specification
; advises a maximum baud rate of 38400 baud.
;
; The testing I performed was with the small test string.
; No tests were made on the actual throughput when transmitting.
; Much testing should be performed to determine if the
; Zoomer PDA is reliable at speeds greater than 38400 baud.
;
subttl  Serial control register
;*******************************************
;* Serial Control Register (R/W), port 51h *
;*******************************************

SERIAL_CONTROL_REG  EQU 51h

; The serial control register controls the number of data bits, stop bits,
; and parity.  This register also enables and disables the receiving and
; transmitting of data and the power mode.
;
; The register is set up as follows:
;   D7   D6   D5   D4   D3   D2   D1   D0
;      STP  EPR  PAE  CLE  TXE  RXE  SM1  SM0
;
ASYNC_MODE  EQU 00000001b
;       SM1 SM0  | Mode select bits
;       ---------+-----------------
;        0   0   | Power down mode
;        0   1   | Asynchronous mode    ** This is the mode you need
;        1   0   | Reserved
;        1   1   | Reserved
;
RXE_BIT EQU 000000100b

;   RXE Receive Enable  - allows the UART to receive data
;         0 = Receive disable
;         1 = Receive enable  - must be set to 1 to receive data
;
TXE_BIT EQU 000001000b

;   TXE Transmit enable - enables the UART to transmit data
;         0 = Transmit disable
;         1 = Transmit enable - must be set to 1 to transmit data
;       This bit MAY be useful for starting and stopping the
;       transmission of characters while using interrupts.  This use
;       has not been tested.
;
CLE_BIT     EQU 00010000b
;
;   CLE Character length - sets the number of data bits to either 7 or 8
;        0 = 7 bits
;        1 = 8 bits
;
PAE_BIT     EQU 00100000b
;
;   PAE Parity Enable   - enables parity checking and generation
;        0 = no parity
;        1 = parity
;
EPR_BIT     EQU 01000000b
;
;   EPR Even Parity - selects even or odd parity if enabled
;                   (see PAE_BIT above)
;        0 = odd parity (if PAE = 1)
;        1 = even parity (if PAE = 1)
;
STP_BIT     EQU 10000000b
;
;   STP Stop bit    - selects 1 or 2 stop bits
;        0 = 1 stop bit
;        1 = 2 stop bits
;
subttl  Serial Status Register
;*********************************************
;* Serial Status Register (R only), port 52h *
;*********************************************

SERIAL_STATUS_REG   EQU 52h
;
; The serial status register returns the status of the transmit and receive
; buffers as well as error information.
;
; The register is set up as follows:
;   D7   D6   D5   D4   D3   D2   D1   D0
;      BRD    0  FER  OER  PER  TEP  RXR  TXR
;
TXR_BIT EQU 00000001b
;   TXR Transmitter ready
;         When this bit is set to 1, you can output the next byte
;         to be transmitted to the Transmitter Holding Register
;         (XMIT_HOLDING_REG, port 56h).
;        0 = first transmit buffer is full
;        1 = first transmit buffer is empty
;
;
;
RXR_BIT EQU 00000010b
;   RXR Receiver ready
;         When this bit is 1, a byte of data has been received and
;         may be read from the Receive buffer register
;         (RECEIVE_BUF_REG, port 57h).
;        0 = no data in receive buffer
;        1 = data in receive buffer
;
TEP_BIT EQU 00000100b
;   TEP Transmit 2nd buffer is empty
;         When this bit is 0, data is still being transmitted,
;         even if TXR is 1.  This bit indicates if data is still
;         in the shift register.  You can output the next byte
;         to be transmitted to the Transmitter Holding Register
;         as long as TXR is 1.
;        0 = first or second transmit buffer is full
;        1 = both first and second transmit buffers are empty
;
PER_BIT EQU 00001000b
;   PER Parity Error
;        A 1 indicates that a parity error occured in the data
;        received.
;        0 = no parity error
;        1 = parity error
;
OER_BIT EQU 00010000b
;   OER Overrun error
;         A 1 indicates that an overrun error occured.  This is when
;         data is received before the previous data has been read by
;         the CPU.  One cause may be running too high of a baud rate
;         or having an overly complicated interrupt handler.
;        0 = no overrun error
;        1 = overrun error
;
FRE_BIT EQU 00100000b
;
;   FRE Framing error
;        0 = no framing error
;        1 = framing error
;
BRD_BIT EQU 10000000b
;   BRD Break detect
;        0 = no break detected
;        1 = break detected
;
; Note:  If this register is written to, the PER, OER, and FRE bits are reset
;    to 0.
;
subttl  Modem Status Register
;********************************************
;* Modem Status Register (R only), port 54h *
;********************************************

MODEM_STATUS_REG    EQU 54h

; The modem status register returns information about different intup
; pins on the serial port.
;
; The register is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;   -    -    -    RI   CD  RDR  CTS  DSR
;
DSR_BIT     EQU 00000001b
;
;   DSR Data Set Ready
;        0 = DSR input pin status is high (1)
;        1 = DSR input pin status is low (0)
;
CTS_BIT     EQU 00000010b
;
;   CTS Clear To Send
;        0 = CTS input pin is high (1)
;        1 = CTS input pin is low (0)
;
RDR_BIT     EQU 00000100b
;
;   RDR Receive Port Data
;        0 = RDR input pin is low (0)
;        1 = RDR input pin is high (1)
;       This bit has two modes based on the SIS bit (bit 2) of the
;       Communications Select Register (port 5Eh).
;
;       Comm Select Reg (5Eh) SIS bit | RDR
;       ------------------------------+-------------------------
;           0             | This bit is connected to
;                         | the RDX port of the
;                         | RS232-C port.
;       ------------------------------+-------------------------
;           1             | This bit is connected to
;                         | the RDX port of the
;                         | infra-red port.
;
CD_BIT      EQU 00001000b
;
;   CD  Carrier Detect
;        0 = CD input pin is high (1)
;        1 = CD input pin is low (0)
;
RI_BIT      EQU 00010000b
;
;   RI  Ring Indicator
;        0 = RI input pin is high (1)
;        1 = RI input pin is low (0)
;
subttl  Modem Control Register
;******************************************
;* Modem Control Register (R/W), port 55h *
;******************************************

MODEM_CONTROL_REG   EQU 55h

; The modem control register allows the RTS, DTR, and TXC lines to be set
;
; The port is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;   -    -    -    -    -   TXC  RTS  DTR
;
RTS_BIT EQU 00000010b

;   RTS Request To Send
;        0 = RTS pin is high (1)
;        1 = RTS pin is low (0)
;
DTR_BIT EQU 00000001b
;   DTR Data Terminal Ready
;        0 = DTR pin is high (1)
;        1 = DTR pin is low (0)
;
TXC_BIT EQU 00000100b

;   TXC Transmit Data Control
;        When TXE bit is set to 0
;         0 = TXC pin is low (0)
;         1 = TXC pin is high (1)
;        When TXE bit is set to 1
;         0 = Break out
;         1 = RS232C transmit data out
;
subttl  Transmitter Holding Register (W only)
;***************************************************
;* Transmitter Holding Register (W only), port 56h *
;***************************************************

XMIT_HOLDING_REG    EQU 56h

; The transmitter holding register holds the value of the next byte to be
; transmitted.  Reading this port will return the value of the last value
; written to it.
;
; Writing to this port will clear the STIE bit (bit 0) of the Serial Interrupt
; Status Register (port 59h).
subttl  Receive buffer register (R only)
;**********************************************
;* Receive Buffer Register (R only), port 57h *
;**********************************************

RECEIVE_BUF_REG EQU 57h

; The receive buffer register contains the value of the last byte received.
; After reading from this port, the RXR bit (bit 1) of the Serial Status
; Register (port 52h) will be reset to 0 and the SRIE bit (bit 1) of the
; Serial Interrupt Clear Register (port 5Ah) will also be cleared.
;
subttl  Serial Interrupt Source Register
;****************************************************
;* Serial Interrupt Source Register (R/W), port 58h *
;****************************************************

SERIAL_INT_SRC_REG  EQU 58h

; The serial interrupt source register allows the UART to generate interrupts
; when certain events occur.
;
; The reguster is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;   -    -    -    -   MST  RLS  RIE  TIE
;
;
TIE_BIT     EQU 00000001b
;
;   TIE Transmit buffer empty Interrupt Enable
;        0 = do not generate an interrupt when transmit buffer is empty
;        1 = generate an interrupt when the transmit buffer is empty
;
RIE_BIT     EQU 00000010b
;
;   RIE Receive buffer full Interrupt Enable
;        0 = do not generate an interrupt when a byte has been received
;        1 = generate an interrupt when a byte has been received
;
RLS_BIT     EQU 00000100b
;
;   RLS Receive Line Status interrupt enable
;        0 = do not generate an interrupt when the line status changes
;        1 = generate an interrupt when the line status changes
;           (Parity error, Overrun error, Framing error)
;
MST_BIT     EQU 00001000b
;
;   MST Modem STatus changed interrupt
;        0 = do not generate an interrupt when the modem status changes
;        1 = generate an interrupt when the modem status changes
;           (CTS, DSR, RI, CD)
;
; Note: TIE, RIE, RLS, MST interrupts do not have a priority level.
;
; To enable interrupts for the serial port, the appropriate bits listed above
; must be set as well as clearing the Interrupt Mask (IM) register (port 20h)
; by ANDing the value in port 20h with 0EFh.  Turning bit 4 of the IM register
; will enable IRQ 4.
subttl  Serial Interrupt Status Register
;*******************************************************
;* Serial Interrupt Status Register (R only), port 59h *
;*******************************************************

SERIAL_INT_STAT_REG EQU 59h

; The serial interrupt status register indicates the condition or conditions
; which caused an interrupt to be generated.  A 1 indicates the condition
; which generated the interrupt.
;
; The register is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;   -    -    -    -   SMST SRLS SRIE STIE
;
STIE_BIT    EQU 00000001b
;
;   STIE    1 = Status of first Transmit buffer empty interrupt source
;           Setting this bit to 1 will cause the UART to generate
;           an interrupt every time TXR is 1.  This is useful for
;           queing data to be transmitted.
;
SRIE_BIT    EQU 00000010b
;
;   SRIE    1 = Status of receive data available interrupt source
;           Setting this bit to 1 will cause an interrupt to be
;           generated whenever data has been received by the UART.
;           This is useful for queing received data in case the
;           data is received faster than it can be processed.
;
SRLS_BIT    EQU 00000100b
;
;   SRLS    1 = Status of receive line status interrupt source
;           This generates an interrupt whenever there
;           is an error detected in the serial transmission,
;           such as a parity error.  An interrupt will also be
;           generated whenever a break signal is received (not tested)
;
SMST_BIT    EQU 00001000b
;
;   SMST    1 = Status of modem status interrupt source
;           This generates an interrupt whenever the state of
;           any of the modem status signals changes, such as CD.
;
;
subttl  Serial Interrupt Clear Register
;******************************************************
;* Serial Interrupt Clear Register (W only), port 5Ah *
;******************************************************

SERIAL_INT_CLR_REG  EQU 5Ah

; The serial interrupt clear register is used to clear the cause of the
; interrupt currently being generated.
;
; The register is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;   -    -    -    -   CMST CRLS CRIE CTIE
;
CTIE_BIT    EQU 00000001b
;
;   CTIE    1 = clear the status of the transmit buffer empty interrupt
;
CRIE_BIT    EQU 00000010b
;
;   CRIE    1 = clear the status of the receive buffer full interrupt
;
CRLS_BIT    EQU 00000100b
;
;   CRLS    1 = clear the status of the line status interrupt
;
CMST_BIT    EQU 00001000b
;
;   CMST    1 = clear the status of the modem status interrupt
;
; After handling all causes of the interrupt, reading the Serial Interrupt
; Status Register (port 59h) and writing the value read to the Serial Interrupt
; Clear Register (port 5Ah) will clear the serial interrupt.
; In addition, the interrupt controller should be told to clear IRQ 4
; by writing the value 10h to the Finish Interrupt service register (port 26h)
;
subttl  Communications Select Register
;**************************************************
;* Communications Select Register (R/W), port 5Eh *
;**************************************************

COMM_SELECT_REG     EQU 5Eh

; The communications select register selects either RS232C or Infra-red I/O.
;
; The register is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;   -    -    -    -    -   SIS  CM1  CM0
;
;   CM0 Carier frequency select of receive data (IR only, ignore for RS232C)
;         This selects the carrier frequency to be used for IR transmission
;         and reception.
;        0 = 150KHz peak
;        1 = 37KHz peak
;
;   CM1 Transmission power select (IR only, ignore for RS232C)
;        0 = low power mode
;        1 = high power mode
;           Note: High power mode should only be used for
;           switch control applications, low power mode should
;           be selected to reduce power consumption.
;
SIS_BIT     EQU 00000100b
SELECT_SERIAL   EQU 00000000b
;
;   SIS Serial Input Select
;         The SIS bit allows either the RS-232C port or the
;         infra-red port to be selected for communication.
;         So far only the RS232 port has been tested.
;        0 = RS232C
;        1 = Infra-red
;
; Note: When using infra-red logic, the SM1 and SM0 bits of the Serial Control
;   Register (port 51h) must be set to 01h (asynchronous mode).
;
subttl  Interrupt Controller Information useful for implementing serial I/O
;************************************
;* Interrupt Controller Information *
;************************************
; The interrupt controller on the Zoomer PDA provides 14 IRQ's, with two
; (12 and 13) being non-maskable and 12 being maskable.
; All of the maskable IRQs may be prioritized.
; All of the IRQ's are masked to interrupts starting at 70h, so IRQ 4, the
; serial IRQ, will generate an interrupt 74h when activated.
;
; All of the IRQs are used in the Zoomer PDA.  The two interrupts of interest
; here are IRQ 4, which is connected to the serial port, and IRQ 8, the
; PCMCIA IRQ (if a PCMCIA fax/modem is used).
;
;*********************************************************
;* Interrupt Mask Register (IM) (R/W), ports 20h and 21h *
;*********************************************************

INTERRUPT_MASK_REG  EQU 20h
INTERRUPT_MASK_REG1 EQU 20h
INTERRUPT_MASK_REG2 EQU 21h

; The interrupt mask register allows interrupts to be masked by setting
; the appropriate bit(s) to 1.  If a bit is set to 1, then the interrupt is
; masked.
;
; The registers are set up as follows:
;         Port    21h       |         Port 20h
; D7   D6   D5   D4   D3   D2   D1   D0 |  D7   D6   D5   D4   D3   D2   D1   D0
;  -    -    -    -  IM11 IM10 IM9  IM8 | IM7  IM6  IM5  IM4  IM3  IM2  IM1  IM0
;
; The only bits of interests are IM8 and IM4, with IM4 being of the most
; interest here.
;
; Clearing IM4 (setting it to 0) allows the serial port to generate interrupts.
; When setting or clearing a bit you should be careful to not change any
; other bits.  The easiest method of enabling IRQ4 is to first read
; port 20h, AND it with 7Fh, and write the new value out to port 20h.
; The following code will accomplish this:
;
;   Code to enable IRQ 4
;
;   in  al,20h
;   and al,7Fh      ; turn off bit 4
;   out 20h,al
;
;   Code to disable IRQ 4
;
;   in  al,20h
;   or  al,10h      ; turn on bit 4
;   out 20h,al
;
;******************************************************************
;* Interrupt In Service (IIS) register (R/W), (Ports 24h and 25h) *
;******************************************************************

INT_IN_SERVICE_REG  EQU 24h
INT_IN_SERVICE_REG1 EQU 24h
INT_IN_SERVICE_REG2 EQU 25h

; The interrupt in service (IIS) register reports if an interrupt
; has been acknowledged by the CPU.  These registers may be used with
; the Finish Interrupt (FI) register.
;
;
; This register is set up as follows:
;       Port 25h                 Port 24h
; D7   D6   D5   D4   D3   D2   D1   D0 |  D7   D6   D5   D4   D3   D2   D1   D0
;  -    -  II13 II12 II11 II10 II9  II8 | II7  II6  II5  II4  II3  II2  II1  II0
;
; II0-II13  Interrupt In service bit
;        0 = IRQ not in service
;        1 = IRQ in service
;
; When writing the IIS register, if a bit was previously 1 and is set to 0
; then that IRQ will be requested.
; If a bit was 0 and is set to 1 then a lower priority interrupt is not
; acknowledged.
;
;**************************************************************
;* Finish Interrupt (FI) register (W only), Ports 26h and 27h *
;**************************************************************

FINISH_INT_REG      EQU 26h
FINISH_INT_REG1     EQU 26h
FINISH_INT_REG2     EQU 27h

; The finish interrupt (FI) register resets the in-service bit for the
; interrupt.
;
; This register is set up as follows:
;       Port 27h                 Port 26h
; D7   D6   D5   D4   D3   D2   D1   D0 |  D7   D6   D5   D4   D3   D2   D1   D0
; AFI   -  FI13 FI12 FI11 FI10 FI9  FI8 | FI7  FI6  FI5  FI4  FI3  FI2  FI1  FI0
;
;   AFI Automatic Finish Interrupt
;        0 = nothing happens
;        1 = highest priority IIS bit is cleared.
;
;   FI0-FI13    Finish IRQn bit
;            0 = not cleared
;            1 = clear IISn bit
;
;   A 1 should be written to either the appropriate bit for the IRQ
;   that was serviced, or a 1 should be written to the AFI bit.
;   This step is usually done just before ending the interrupt service
;   routine via IRET.  This tells the interrupt controller that the interrupt
;   is finished and another may be generated.
;
;*******************************************************
;* Interrupt Priority Register (R/W), Ports 30h to 3Bh *
;*******************************************************
;
;
INT_PRIORITY_REG    EQU 30h
INT_0_PRIORITY_REG  EQU 30h
INT_1_PRIORITY_REG  EQU 31h
INT_2_PRIORITY_REG  EQU 32h
INT_3_PRIORITY_REG  EQU 33h
INT_4_PRIORITY_REG  EQU 34h
INT_5_PRIORITY_REG  EQU 35h
INT_6_PRIORITY_REG  EQU 36h
INT_7_PRIORITY_REG  EQU 37h
INT_8_PRIORITY_REG  EQU 38h
INT_9_PRIORITY_REG  EQU 39h
INT_10_PRIORITY_REG EQU 3Ah
INT_11_PRIORITY_REG EQU 3Bh

; The interrupt priority register allows priorities to be assigned to different
; IRQs.  IRQs 12 and 13 (the non-maskable interrupts) are the highest priority
; and cannot be changed.  IRQs 0 to 11 are programmable.
;
; Port 30 holds the priority of IRQ 0.  The priority can be set from 0 to 11,
;   where 0 is the highest priority and 11 is the lowest.  Only the lower 4
;   bits are used.  Levels set higher than 11 are treated the same as 11.
; The priority ports for IRQ's 0 to 11 are listed as follows:
;
;   IRQ   | Port
;   ------+-----
;   IRQ0  | 30h
;   IRQ1  | 31h
;   IRQ2  | 32h
;   IRQ3  | 33h
;   IRQ4  | 34h
;   IRQ5  | 35h
;   IRQ6  | 36h
;   IRQ7  | 37h
;   IRQ8  | 38h
;   IRQ9  | 39h
;   IRQ10 | 3Ah
;   IRQ11 | 3Bh
;
; If each register is set to the same priority level, then the interrupt
; requests will be set to the same level.
; Interrupt level n = URQk + IRQl
;
; These registers cannot be set while an interrupt is in service.
;
; I have not found any reason yet to change the interrupt priorities.
;
subttl  Power Control Register
;******************************
;* Power Management Registers *
;******************************
;***************************************************
;* RS232C Output Control Register (R/W), port 0DEh *
;***************************************************

RS232C_OUTPUT_CONTROL_REG   EQU 0DEh

; The RS232 output control register enables and disables the output of the
; RS232 port.
;
; The register is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;      ROC    -    -    -    -    -    -    -
;
RS232C_OUTPUT_DISABLE   EQU 00000000b
RS232C_OUTPUT_ENABLE    EQU 10000000b
;
;   ROC RS232C Output Control bit
;         I think this bit is used to control whether the outputs
;         of the UART make it to the outside world.  This would
;         help conserve power, but makes the serial port useless
;         until it is enabled.
;
;         It's probably a good idea to disable the output when the
;         serial port is not being used.
;        0 = RS232C output disabled
;        1 = RS232C output enabled
;
;*****************************************************************
;* RS232C Line Driver/Receiver Control Register (R/W), port 0DFh *
;*****************************************************************

RS232C_LINE_DRIVER_CONTROL_REG  EQU 0DFh

; The RS232C line driver/receiver control register controls the power to the
; RS232C port and what mode it is in.
;
; The register is set up as follows:
;
;   D7   D6   D5   D4   D3   D2   D1   D0
;    -    -    -    -    -  RSE  RSS  RSP
;
RS232C_POWER_ON EQU 00000001b
;
;   RSP RS232C power supply switch
;         I think this bit controls the power to the UART.
;         Turning this off may reset all of the UART registers.
;                 I have not tested this theory.
;        0 = RS232C power off
;        1 = RS232C power on
;
RS232C_ACTIVE   EQU 00000110b
;
;   RSE RSS RS232C device control bits
;    0   0   Standby mode       ; I think this just puts the UART into a
;                   ; low power state.
;    0   1   3-state mode (not used)
;        1   0   Wake up mode           ;
;    1   1   Active mode        ; This is the mode where it's all happening
;
subttl  Setting communications conditions
;****************************************************
;* Note on how to set the communications conditions *
;****************************************************
; To set the communications conditions, the following steps should be
; taken in order.
;
;   1. Set the power control registers
;   2. Set the communication select register (port 005Eh)
;   3. Set the conditions of communication (baud, parity, etc.)
;   4. Set the conditions for interrupts
;   5. Set the bits (TXE and/or RXE) of the serial control register.
;      The condition of communication must be set during TXE=RXE=0.
;
page
;***************************************************************
; Beginning of our code
.code
.186
JUMPS
	ORG 0100h
start:  jmp start1

int_addr    dd  ?   ; old interrupt handler address

xmit_queue      db  4096    dup('T')    ; transmitting queue
xmit_queue_start    dw  0   ; head of transmitter queue
xmit_queue_end      dw  0   ; tail of queue
xmit_queue_size     dw  4095    ; total number of bytes available in queue
xmit_queue_count    dw  0   ; Number of bytes stored in queue
					;  (must always be < xmit_queue_size)

recv_queue      db  4096    dup('R')    ; receiving queue
recv_queue_start    dw  0   ; head of queue
recv_queue_end      dw  0   ; tail of queue
recv_queue_size     dw  4095    ; total number of bytes in queue
recv_queue_count    dw  0   ; number of bytes stored in queue

; Strings used
hexdata db  '0123456789ABCDEF'  ; data used by display_hex

test_string     db  'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz',13,10,0
modem_status_str    db  'Modem status changed.',13,10,0
line_status_str     db  'Line status changed.',13,10,0
init_str        db  'Zoomer PDA Serial sample program',13,0

;Debugging aids

;********************************************************************
; display_hex proc
;
; This proc will display the value in AL as a two-character hex value
; The ah, bx, cl, and dx registers are destroyed.  This routine has NOT
; been tested.
;********************************************************************
display_hex proc
	mov cl,al
	xor dx,dx
	mov bx,offset hexdata
	mov ah,2
	mov dl,al
	shr dl,4
	add bx,dx
	mov dl,byte ptr cs:[bx]
	int 21h
	mov dl,cl
	mov ah,2
	and dl,0Fh
	mov bx,offset hexdata
	add bx,dx
	mov dl,byte ptr cs:[bx]
	int 21h
	ret
display_hex ENDP

subttl  display_char
page
;********************************************************************
;* display_char                             *
;*  Displays a character using int 21h function 2           *
;********************************************************************
;*  Input:                              *
;*  AL - ASCII code of character to be displayed            *
;*  Output:                             *
;*  none                                *
;*  Registers modified:                         *
;*  none                                *
;********************************************************************
display_char    proc
	push    ax
	push    dx
	mov ah,2
	mov dl,al
	int 21h
	pop dx
	pop ax
	ret
display_char    endp

subttl  display_str
page
;********************************************************************
;* display_str   MACRO                          *
;*  Displays a '$' terminated string using int 21h function 9   *
;********************************************************************
;*  Input:                              *
;*  DS:DX - address of '$' terminated string            *
;*  Output:                             *
;*  none                                *
;*  Registers modified:                         *
;*  none                                *
;********************************************************************
display_str MACRO   string
	push    ax
	push    dx
	mov dx, offset string
	mov ah,9
	int 21h
	pop dx
	pop ax
	ENDM

subttl  show_value_as_bitmap
page
;********************************************************************
;* show_value_as_bitmap MACRO                       *
;*  Displays value at the coordinates specified by xpos and     *
;*  ypos.  This code displays the value as a series of 8 pixels *
;*  on the display, with the MSB being the leftmost pixel.      *
;*  This macro may be used anywhere, including inside of an     *
;*  interrupt service routine.                  *
;*  Having the value '55h' displayed either above or below will *
;*  make it much easier to determine the value on the screen.   *
;*  Darkened pixels indicate '1' bits and light pixels indicate *
;*  '0' bits.                           *
;********************************************************************
;*  Inputs:                             *
;*  xpos - x position to display.  0 is the left side, 1 is     *
;*         indented 8 pixels, etc.                  *
;*         Valid values are 0 to 31.  Value not checked.        *
;*  ypos - y position on display with 0 being the top and 319   *
;*         being the bottom.                    *
;*  value - the value to be displayed.  value may be from the   *
;*      following registers or just a plain byte value:     *
;*      AH, BL, BH, CL, CH                  *
;*      All other 8-bit registers are destroyed         *
;*  Output:                             *
;*  none                                *
;*  Registers modified:                         *
;*  none                                *
;********************************************************************
show_value_as_bitmap    MACRO   value, xpos, ypos
	push    ax
	push    es
	push    si
	mov si,0D800h   ; segment of video memory in the Zoomer PDA
	mov es,si

	mov si,ypos
	shl si,6        ; y * 64
	add si,xpos
	mov al,value
	mov byte ptr es:[si],al
	pop si
	pop es
	pop ax
ENDM

;********************************************************************
;* transmit_wait macro
;*
;* Waits until transmit buffer is empty
;*
;* Registers used:  none
;* Registers destroyed: none
transmit_wait   MACRO
local   transmitter_full
	push    ax
transmitter_full:
	in  al,SERIAL_STATUS_REG
	test    al,TXR_BIT
	jz  transmitter_full
	pop ax
	ENDM

;********************************************************************
;* clrscr
;*
;* Clears the screen on the Zoomer PDA and places the cursor in the upper
;* left hand corner.
;*
;* Registers destroyed: ax and dx, others not tested.
clrscr  PROC
	push    ds
	push    es
	mov ax,0600h
	int 10h
	mov ah,2
	xor dx,dx
	int 10h
	pop es
	pop ds
	ret
clrscr  ENDP

;********************************************************************
;* print_asciiz
;*
;* Prints a null-terminated ASCII string pointed to DS:SI via
;* int 10h function 0Eh
;*
;* Registers destroyed: none
print_asciiz    PROC
	push    cx
	push    si
	push    ds
	push    es
	mov cx,1
print_asciiz_start:
	lodsb
	cmp al,0
	je  print_asciiz_done
	mov ah,0Eh
	int 10h
	jmp print_asciiz_start

print_asciiz_done:
	pop es
	pop ds
	pop si
	pop cx
	ret
print_asciiz    ENDP

;********************************************************************
;* xmit_char
;*
;* Transmits the character in the AL register.
;*
;* If the queue is empty and the transmit buffer is empty, the character
;* is immediately transmitted.
;* Otherwise, the character is added to the queue and transmitting interrupts
;* are enabled.
;*
;* Registers destroyed: ax, bx, cx, dx
;*
;* Returns: Carry flag - when queue is full
;*
;* This code can be called inside of an ISR since it never explicitly enables
;* interrupts, instead it just returns the interrupt enable flag to its
;* previous state.
xmit_char   PROC
	mov cl,al       ; copy character to transmit into cl
	pushf           ; save flags
	mov dx, word ptr [xmit_queue_count] ; get size of queue
	cmp dx,0        ; is the queue empty?
	je  @@queue_empty   ; yes

	cmp dx, word ptr [xmit_queue_size]  ; is the queue full?
	jb  @@queue_not_full        ; no

	popf            ; yes, restore flags
	stc         ; set carry to mark an error
	ret         ; return error

@@queue_empty:      ; the queue is empty at this point

	in  al,SERIAL_STATUS_REG    ; is the transmit buffer full?
	test    al,TXR_BIT
	jz  @@add_to_queue      ; yes, so add character to queue

					; no, so we can just bypass the queue
					; and transmit the character
					; immediately
	mov al,cl
	out XMIT_HOLDING_REG,al ; transmit character immediately
	popf                ; restore flags
	clc             ; no carry indicates no error
	ret             ; return

@@queue_not_full:
@@add_to_queue:
	cli             ; disable interrupts at this point
	mov bx, [xmit_queue_end]    ; get the end or tail of queue
	inc bx          ; increment tail of queue
	cmp bx, [xmit_queue_size]   ; do we wrap?
	jb  @@no_xmit_wrap      ; no
	xor bx,bx           ; yes, go back to beginning
@@no_xmit_wrap:
	mov [xmit_queue_end],bx ; update end of queue
	mov byte ptr [xmit_queue+bx],cl ; store character in queue
	inc word ptr [xmit_queue_count] ; increment count

	popf                ; restore flags to enable interrupts
					; if they were enabled before xmit_char
					; was called

	in  al,SERIAL_INT_SRC_REG   ; enable transmit interrupts
	or  al,TIE_BIT
	out SERIAL_INT_SRC_REG,al

	clc             ; no errors
	ret
xmit_char   ENDP


;********************************************************************
;* sendstring
;*
;* Sendstring transmits a null terminated string through the serial port
;* The string is pointed to by DS:SI
;*
;* Note that if the queue is full we keep trying for a while
;* (A few seconds on the Zoomer PDA) until either we give up or a slot in the
;* queue opens up.
;*
;* registers destroyed: none
;* Returns: Carry Flag - set if error occured
sendstring  PROC
	pusha           ; save all registers
send_string_loop:
	lodsb               ; get character in string
	cmp al,0            ; end of string?
	je  send_string_done    ; yes, done
	mov cx,0FFFFh       ; number of times to try
try_again:
	dec cx
	jz  send_string_error   ; jump if can't send after trying cx times

	push    ax      ; save character
	call    xmit_char   ; no, transmit a character
	pop ax      ; restore character in case we must try again
	jc  try_again   ; an error occured (full queue), so try again

	jmp short send_string_loop  ; check next character position

send_string_done:
	popa        ; restore all registers
	clc     ; no errors
	ret

send_string_error:
	popa        ; restore all registers
	stc     ; report error transmitting (queue never emptied)
			; (interrupt not responding or working)
	ret
sendstring  ENDP



;*****************************************************************
start1:         ; beginning of initialization code
	mov ax,cs   ; set data and extra segments
	mov ds,ax
	mov es,ax

	; Set TXE and RXE bits to 0  (no transmitting or receiving allowed yet)
	mov al,ASYNC_MODE       ; set to asynchronous mode
	out SERIAL_CONTROL_REG,al   ; (where SERIAL_CONTROL_REG is initialized)

	; 1. Set the power control registers

	mov al,RS232C_OUTPUT_ENABLE     ; enable RS232C output
	out RS232C_OUTPUT_CONTROL_REG, al   ;

	mov al,RS232C_POWER_ON OR RS232C_ACTIVE ; make RS232C active and
	out RS232C_LINE_DRIVER_CONTROL_REG, al  ; powered on

	; 2. Set communications select register
	mov al,SELECT_SERIAL
	out COMM_SELECT_REG, al ; select RS232 output, not infra-red

	; 3. Set condition of communications
	mov al, Z19200_BAUD      ; set baud rate to 19200 baud
	out BAUD_RATE_REG, al

	in  al,SERIAL_CONTROL_REG   ; (initialized above)
	or  al, CLE_BIT     ; 8 data bits, 1 stop, no parity
	and al, NOT (PAE_BIT OR EPR_BIT OR STP_BIT)
	out SERIAL_CONTROL_REG, al

	mov al,DTR_BIT+TXC_BIT  ; Set DTR
	out MODEM_CONTROL_REG, al   ; output RS232 data (bit 2)

	mov ax,3574h    ; get old interrupt handler address
	int 21h     ; (This may not really be necessary)
	mov word ptr cs:[int_addr+2],es ; save it
	mov word ptr cs:[int_addr],bx

	lea dx,int_handler  ; get address of our interrupt handler
	mov ax,2574h
	int 21h     ; install our interrupt handler

	cli         ; disable interrupts while setting up
				; masks
	; 4. Set condition of interrupt
	in  al,INTERRUPT_MASK_REG
	and al,0EFh     ; enable IRQ 4 (IRQ of com port, int handler = 74h)
	out INTERRUPT_MASK_REG,al   ; unmask interrupt from controller

	; 5. Set bits (TXE and/or RXE) of serial control register.
	in  al,SERIAL_CONTROL_REG
	or  al, RXE_BIT OR TXE_BIT OR ASYNC_MODE
	out SERIAL_CONTROL_REG, al  ; allow UART to send and receive


	mov al,TIE_BIT OR RIE_BIT OR RLS_BIT OR MST_BIT ; enable all serial interrupts
	out SERIAL_INT_SRC_REG,al

	mov al,CTIE_BIT OR CRIE_BIT OR CRLS_BIT OR CMST_BIT
	out SERIAL_INT_CLR_REG,al   ; clear any pending serial interrupts
					; just in case.

	sti             ; enable interrupts
	mov si, offset test_string
	call    sendstring      ; send a test string to serial port
	jnc no_error

	mov al,'x'          ; if an error is reported, send 'x'
	out XMIT_HOLDING_REG,al

no_error:

	xor ax,ax       ; wait for keystroke (zkeypad must be loaded)
	int 16h

	cli         ; disable interrupts
	in  al,INTERRUPT_MASK_REG1  ; mask off our interrupt
	or  al,10h      ; actually we should restore the state before we installed
	out INTERRUPT_MASK_REG1,al  ; mask off IRQ 4
	mov al,0            ; disable all serial interrupts
	out SERIAL_INT_SRC_REG,al   ; we should actually restore this value

	mov al,CTIE_BIT OR CRIE_BIT OR CRLS_BIT OR CMST_BIT
	out SERIAL_INT_CLR_REG,al   ; clear any pending serial interrupts
					; just in case

	mov es,word ptr cs:[int_addr+2]
	mov dx, word ptr cs:[int_addr]
	mov ax,2574h        ; restore old IRQ handler
	int 21h
	sti             ; enable interrupts again

	mov ax,04c00h       ; terminate program
	int 21h

subttl  Serial ISR
page
;********************************************************************
;* Serial interrupt handler for the Zoomer PDA
;*
;* The serial port of the Zoomer PDA is mapped to IRQ 4, which is set up to
;* interrupt vector 74h.
int_handler proc    far
	pusha           ; save all registers
	push    ds      ; save segment registers
	push    es

	mov ax,cs
	mov ds,ax       ; set data segment and extra segment
	mov es,ax

beginning_of_serial_int:
	; jump to appropriate code for interrupt handler

	in  al,SERIAL_INT_STAT_REG  ; is this a valid serial interrupt?
	test    al,STIE_BIT OR SRIE_BIT OR SRLS_BIT OR SMST_BIT
	jz  end_of_serial_int   ; no, exit

					; yes
	test    al,SRIE_BIT     ; was a character received?
	jz  receiver_data_not_available ; no

receiver_data_available:        ; a character is available to be read
; Note that the queue code here has NOT been tested in any way.
; You have been warned!
	in  al,RECEIVE_BUF_REG  ; read character from com port

	mov bx,[recv_queue_end] ; get tail of queue
	inc bx
	cmp bx,[recv_queue_size]    ; do we wrap around the array?
	jb  @@rxq_no_wrap       ; no
	xor bx,bx           ; yes
@@rxq_no_wrap:
	mov [recv_queue_end],bx ; set new tail
	mov byte ptr [recv_queue+bx],al ; put character in queue
	inc word ptr [recv_queue_count] ; increment number of characters in queue

	transmit_wait
	out XMIT_HOLDING_REG,al ; transmit character that was read

@@clear_receive_int:
	mov al,CRIE_BIT ; clear receive data available interrupt
	out SERIAL_INT_CLR_REG,al

receiver_data_not_available:

;**************************** Transmit data code starts here ********
; All the queue code here has been tested and shown to work properly
	in  al,SERIAL_INT_STAT_REG      ; is the transmit buffer empty?
	test    al,TXR_BIT
	jz  transmit_buffer_not_empty   ; jump if data in buffer

transmit_buffer_empty:      ; if transmit buffer is empty, check queue
	mov cx, [xmit_queue_count]
	cmp cx,0            ; is queue empty?
	je  @@no_transmit       ; yes, do nothing

	mov bx,[xmit_queue_start]        ; get start of queue
	inc bx
	cmp bx,[xmit_queue_size]        ; get end of queue
	jb  @@no_xmit_wrap
	xor bx,bx               ; wrap back to beginning of queue

@@no_xmit_wrap:
	mov [xmit_queue_start],bx       ; set new beginning of queue
	mov ah, byte ptr [xmit_queue+bx]    ; get byte from queue
	mov al,ah
	out XMIT_HOLDING_REG,al     ; transmit character from queue
	dec word ptr [xmit_queue_count] ; decrement count of characters
						; left in queue
;    mov     al,CTIE_BIT
;    out     SERIAL_INT_CLR_REG,al

	jmp transmit_buffer_not_empty   ; done with transmit part

@@no_transmit:         ; no more data to send

	in  al,SERIAL_INT_SRC_REG
	and al,NOT TIE_BIT          ; disable TXE interrupt
	out SERIAL_INT_SRC_REG,al

transmit_buffer_not_empty:

	in  al,SERIAL_INT_STAT_REG  ; check to see if RECEIVE line status
	test    al,SRLS_BIT     ; caused the interrupt
	jz  no_receive_line_status  ; jump if receive line status not changed

receive_line_status:
; This has not been tested fully
; I think this will work
	transmit_wait           ; wait for transmitter to flush
					; bad thing to do in ISR!!
	mov al,'l'
	out XMIT_HOLDING_REG,al
	mov si,offset line_status_str
;    call    sendstring (Not without further testing)
	call    print_asciiz
	mov al,CRLS_BIT
	out SERIAL_INT_CLR_REG,al   ; tell UART we handled it
no_receive_line_status:

	in  al,SERIAL_INT_STAT_REG
	test    al,SMST_BIT         ; did the modem status change?
	jz  no_modem_status_interrupt   ; jump if no status change

modem_status_interrupt:
;This has not been tested at all!  I have no easy way to change the
;modem status lines without a breakout box or something other than a standard
;communications program at the other end.
;I think this should work.
	transmit_wait           ; wait for transmitter to flush
					; bad thing to do in ISR!!
	mov al,'m'
	out XMIT_HOLDING_REG,al
	mov si,offset modem_status_str  ; transmit a string indicating
;    call    sendstring          ; that a modem status changed
	call    print_asciiz


	mov al,CMST_BIT         ; clear a modem status interrupt
	out SERIAL_INT_CLR_REG,al
no_modem_status_interrupt:

end_of_serial_int:
	in  al,SERIAL_INT_STAT_REG  ; clear all active interrupts in the
	out SERIAL_INT_CLR_REG,al   ; serial interrupt clear register

	mov al,10h          ; tell interrupt controller that
	out FINISH_INT_REG1,al  ; IRQ 4 is finished

	pop es          ; restore all registers
	pop ds
	popa
	iret                ; return from interrupt
int_handler endp

ends

end start


